import os
import time
import random
import shutil
import zipfile

from hashlib import md5

from DIRAC import gConfig, gLogger
from DIRAC.Core.Utilities import Time
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup

from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen


class FileCatalogHandler(WebHandler):

    AUTH_PROPS = "authenticated"

    def initializeRequest(self):
        self.user = self.getUserName()
        self.group = self.getUserGroup()
        self.vo = getVOForGroup(self.group)
        self.fc = FileCatalog(vo=self.vo)

    @asyncGen
    def web_getSelectedFiles(self):
        """Method to get the selected file(s)"""
        gLogger.always("getSelectedFiles: incoming arguments", repr(self.request.arguments))

        # First pass: download files and check for the success
        if "archivePath" not in self.request.arguments:
            tmpdir = "/tmp/" + str(time.time()) + str(random.random())
            dataMgr = DataManager(vo=self.vo)
            lfnStr = self.get_argument("path")
            if not os.path.isdir(tmpdir):
                os.makedirs(tmpdir)
            os.chdir(tmpdir)
            for lfn in lfnStr.split(","):
                gLogger.always("Data manager get file %s" % lfn)
                last_slash = lfn.rfind("/")
                pos_relative = lfn.find("/")
                pos_relative = lfn.find("/", pos_relative + 1)
                pos_relative = lfn.find("/", pos_relative + 1)
                pos_relative = pos_relative
                pathInZip = lfn[pos_relative:last_slash]
                tmpPathInZip = tmpdir + pathInZip
                gLogger.always("path in zip %s" % tmpPathInZip)
                if not os.path.isdir(tmpPathInZip):
                    os.makedirs(tmpPathInZip)
                result = dataMgr.getFile(str(lfn), destinationDir=str(tmpPathInZip))
                if not result["OK"]:
                    gLogger.error("Error getting while getting files", result["Message"])
                    self.finish({"success": "false", "error": result["Message"], "lfn": lfn})
                    shutil.rmtree(tmpdir)
                    return

            # make zip file
            zipname = tmpdir.split("/")[-1] + ".zip"
            archivePath = "/tmp/" + zipname
            zFile = zipfile.ZipFile(archivePath, "w")
            gLogger.always("zip file %s" % archivePath)
            gLogger.always("start walk in tmpdir %s" % tmpdir)
            for absolutePath, dirs, files in os.walk(tmpdir):
                gLogger.always("absolute path %s" % absolutePath)
                gLogger.always("files %s" % files)
                for filename in files:
                    # relative path form tmpdir current chdir
                    pos_relative = absolutePath.find("/")
                    pos_relative = absolutePath.find("/", pos_relative + 1)
                    pos_relative = absolutePath.find("/", pos_relative + 1)
                    pos_relative = absolutePath.find("/", pos_relative + 1)
                    pos_relative = pos_relative + 1
                    relativePath = absolutePath[pos_relative:]
                    gLogger.always("relativePath %s, file %s" % (relativePath, filename))
                    zFile.write(os.path.join(absolutePath, filename))
            zFile.close()
            shutil.rmtree(tmpdir)
            self.finish({"success": "true", "archivePath": archivePath})

        else:
            # Second pass: deliver the requested archive
            archivePath = self.get_argument("archivePath")
            # read zip file
            with open(archivePath, "rb") as archive:
                data = archive.read()
            # cleanup
            os.remove(archivePath)

            self.set_header("Content-type", "text/plain")
            self.set_header("Content-Length", len(data))
            self.set_header("Content-Disposition", 'attachment; filename="' + os.path.basename(archivePath))

            self.finish(data)

    @asyncGen
    def web_getMetadataFields(self):
        """Method to read all the available fields possible for defining a query"""
        self.L_NUMBER = 0
        self.S_NUMBER = 0
        result = yield self.threadTask(self.fc.getMetadataFields)
        gLogger.debug("request: %s" % result)
        if not result["OK"]:
            gLogger.error("getSelectorGrid: %s" % result["Message"])
            self.finish({"success": "false", "error": result["Message"]})
            return
        result = result["Value"]
        callback = {}
        if "FileMetaFields" not in result:
            error = "Service response has no FileMetaFields key"
            gLogger.error("getSelectorGrid: %s" % error)
            self.finish({"success": "false", "error": error})
            return
        if "DirectoryMetaFields" not in result:
            error = "Service response has no DirectoryMetaFields key"
            gLogger.error("getSelectorGrid: %s" % error)
            self.finish({"success": "false", "error": error})
            return
        filemeta = result["FileMetaFields"]
        if len(filemeta) > 0:
            for key, value in filemeta.items():
                callback[key] = "label"
        gLogger.debug("getSelectorGrid: FileMetaFields callback %s" % callback)
        dirmeta = result["DirectoryMetaFields"]
        if len(dirmeta) > 0:
            for key, value in dirmeta.items():
                callback[key] = value.lower()
        gLogger.debug("getSelectorGrid: Resulting callback %s" % callback)
        self.finish({"success": "true", "result": callback})

    @asyncGen
    def web_getQueryData(self):
        """Method to read all the available options for a metadata field"""
        try:
            compat = {}
            for key in self.request.arguments:
                parts = key.split(".")
                if len(parts) != 3:
                    continue
                _, name, sign = parts

                if not name:
                    continue

                value = self.get_argument(key).split("|")

                # check existence of the 'name' section
                if name not in compat:
                    compat[name] = {}

                # check existence of the 'sign' section
                if sign not in compat[name]:
                    if value[0] == "v":
                        compat[name][sign] = ""
                    elif value[0] == "s":
                        compat[name][sign] = []

                if value[0] == "v":
                    compat[name][sign] = value[1]
                elif value[0] == "s":
                    compat[name][sign] += value[1].split(":::")

        except Exception:
            self.finish({"success": "false", "error": "Metadata query error"})
            return

        path = "/"

        if "path" in self.request.arguments:
            path = self.get_argument("path")

        gLogger.always(compat)

        result = yield self.threadTask(self.fc.getCompatibleMetadata, compat, path)
        gLogger.always(result)

        if not result["OK"]:
            self.finish({"success": "false", "error": result["Message"]})
            return

        self.finish({"success": "true", "result": result["Value"]})

    @asyncGen
    def web_getFilesData(self):
        req = self.__request()
        gLogger.always(req)
        gLogger.debug("submit: incoming request %s" % req)
        result = yield self.threadTask(
            self.fc.findFilesByMetadataWeb, req["selection"], req["path"], self.S_NUMBER, self.L_NUMBER
        )
        gLogger.debug("submit: result of findFilesByMetadataDetailed %s" % result)
        if not result["OK"]:
            gLogger.error("submit: %s" % result["Message"])
            self.finish({"success": "false", "error": result["Message"]})
            return
        result = result["Value"]

        if not len(result) > 0:
            self.finish({"success": "true", "result": [], "total": 0, "date": "-"})
            return

        total = result["TotalRecords"]
        result = result["Records"]

        callback = list()
        for key, value in result.items():

            size = ""
            if "Size" in value:
                size = value["Size"]

            date = ""
            if "CreationDate" in value:
                date = str(value["CreationDate"])

            meta = ""
            if "Metadata" in value:
                m = value["Metadata"]
                meta = "; ".join(["%s: %s" % (i, j) for (i, j) in m.items()])

            dirnameList = key.split("/")
            dirname = "/".join(dirnameList[: len(dirnameList) - 1])
            filename = dirnameList[len(dirnameList) - 1 :]

            callback.append(
                {
                    "fullfilename": key,
                    "dirname": dirname,
                    "filename": filename,
                    "date": date,
                    "size": size,
                    "metadata": meta,
                }
            )
        timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")
        self.finish({"success": "true", "result": callback, "total": total, "date": timestamp})

    def __request(self):
        req = {"selection": {}, "path": "/"}

        self.L_NUMBER = 25
        if "limit" in self.request.arguments and len(self.get_argument("limit")) > 0:
            self.L_NUMBER = int(self.get_argument("limit"))

        self.S_NUMBER = 0
        if "start" in self.request.arguments and len(self.get_argument("start")) > 0:
            self.S_NUMBER = int(self.get_argument("start"))

        result = gConfig.getOption("/WebApp/ListSeparator")
        if result["OK"]:
            separator = result["Value"]
        else:
            separator = ":::"

        result = self.fc.getMetadataFields()
        gLogger.debug("request: %s" % result)

        if not result["OK"]:
            gLogger.error("request: %s" % result["Message"])
            return req
        result = result["Value"]

        if "FileMetaFields" not in result:
            error = "Service response has no FileMetaFields key. Return empty dict"
            gLogger.error("request: %s" % error)
            return req

        if "DirectoryMetaFields" not in result:
            error = "Service response has no DirectoryMetaFields key. Return empty dict"
            gLogger.error("request: %s" % error)
            return req

        filemeta = result["FileMetaFields"]
        dirmeta = result["DirectoryMetaFields"]

        meta = []
        for key, value in dirmeta.items():
            meta.append(key)

        gLogger.always("request: metafields: %s " % meta)

        for param in self.request.arguments:
            tmp = param.split(".")
            if len(tmp) != 3:
                continue
            _, name, logic = tmp
            value = self.get_argument(param).split("|")
            if logic not in ["in", "nin", "=", "!=", ">=", "<=", ">", "<"]:
                gLogger.always("Operand '%s' is not supported " % logic)
                continue

            if name in meta:
                # check existence of the 'name' section
                if name not in req["selection"]:
                    req["selection"][name] = {}

                # check existence of the 'sign' section
                if logic not in req["selection"][name]:
                    if value[0] == "v":
                        req["selection"][name][logic] = ""
                    elif value[0] == "s":
                        req["selection"][name][logic] = []

                if value[0] == "v":
                    req["selection"][name][logic] = value[1]
                elif value[0] == "s":
                    req["selection"][name][logic] += value[1].split(":::")
        if "path" in self.request.arguments:
            req["path"] = self.get_argument("path")
        gLogger.always("REQ: ", req)
        return req

    def __request_file(self):
        req = {"selection": {}, "path": "/"}

        separator = ":::"

        result = self.fc.getMetadataFields()
        gLogger.debug("request: %s" % result)

        if not result["OK"]:
            gLogger.error("request: %s" % result["Message"])
            return req
        result = result["Value"]

        if "FileMetaFields" not in result:
            error = "Service response has no FileMetaFields key. Return empty dict"
            gLogger.error("request: %s" % error)
            return req

        if "DirectoryMetaFields" not in result:
            error = "Service response has no DirectoryMetaFields key. Return empty dict"
            gLogger.error("request: %s" % error)
            return req

        filemeta = result["FileMetaFields"]
        dirmeta = result["DirectoryMetaFields"]

        meta = []
        for key, value in dirmeta.items():
            meta.append(key)

        gLogger.always("request: metafields: %s " % meta)

        selectionElems = self.get_argument("selection").split("<|>")

        gLogger.always("request: THISSSS %s " % self.get_argument("selection"))

        for param in selectionElems:

            tmp = str(param).split("|")

            if len(tmp) != 4:
                continue

            name = tmp[0]
            logic = tmp[1]

            if logic not in ["in", "nin", "=", "!=", ">=", "<=", ">", "<"]:
                gLogger.always("Operand '%s' is not supported " % logic)
                continue

            if name in meta:
                # check existence of the 'name' section
                if name not in req["selection"]:
                    req["selection"][name] = dict()

                # check existence of the 'sign' section
                if logic not in req["selection"][name]:
                    if tmp[2] == "v":
                        req["selection"][name][logic] = ""
                    elif tmp[2] == "s":
                        req["selection"][name][logic] = []

                if tmp[2] == "v":
                    req["selection"][name][logic] = tmp[3]
                elif tmp[2] == "s":
                    req["selection"][name][logic] += tmp[3].split(":::")
        if "path" in self.request.arguments:
            req["path"] = self.get_argument("path")
        gLogger.always("REQ: ", req)
        return req

    @asyncGen
    def web_getMetadataFilesInFile(self):
        self.set_header("Content-type", "text/plain")
        self.set_header("Content-Disposition", 'attachment; filename="error.txt"')
        req = self.__request_file()
        gLogger.always(req)
        gLogger.debug("submit: incoming request %s" % req)
        result = yield self.threadTask(self.fc.findFilesByMetadata, req["selection"], req["path"])

        if not result["OK"]:
            gLogger.error("submit: %s" % result["Message"])
            self.finish({"success": "false", "error": result["Message"]})
            return

        result = result["Value"]
        retStrLines = []

        if len(result) > 0:
            for fileName in result:
                retStrLines.append(fileName)

        strData = "\n".join(retStrLines)

        self.set_header("Content-type", "text/plain")
        self.set_header("Content-Disposition", 'attachment; filename="%s.txt"' % md5(str(req).encode()).hexdigest())
        self.set_header("Content-Length", len(strData))
        self.finish(strData)

    @asyncGen
    def web_getSubnodeFiles(self):
        path = self.get_argument("path")

        result = yield self.threadTask(self.fc.listDirectory, path, False)
        if not result["OK"]:
            gLogger.error("submit: %s" % result["Message"])
            self.finish({"success": "false", "error": result["Message"]})
            return
        filesData = result["Value"]["Successful"][path]["Files"]
        dirData = result["Value"]["Successful"][path]["SubDirs"]

        retData = []

        for entryName in dirData:
            nodeDef = {"text": entryName.split("/")[-1]}
            nodeDef["leaf"] = False
            nodeDef["expanded"] = False
            retData.append(nodeDef)

        for entryName in filesData:
            nodeDef = {"text": entryName.split("/")[-1]}
            nodeDef["leaf"] = True
            retData.append(nodeDef)

        retData = sorted(retData, key=lambda node: node["text"].upper())

        self.finish({"success": "true", "nodes": retData})



""" Basic modules for loading handlers
"""

import os
import collections

from DIRAC import S_OK, S_ERROR, rootPath, gLogger
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
from DIRAC.Core.Utilities.Extensions import extensionsByPriority, getExtensionMetadata

from WebAppDIRAC.Lib import Conf
from WebAppDIRAC.Lib.WebHandler import WebHandler, _WebHandler
from WebAppDIRAC.Core.StaticHandler import StaticHandler


class HandlerMgr(metaclass=DIRACSingleton):
    """This class prepares portal application handlers and forms their routing using regular expressions,
    see https://docs.python.org/3/library/re.html
    """

    def __init__(self, handlersLocation, baseURL="/"):
        """Constructor

        :param str handlersLocation: handlers location
        :param str baseURL: base URL
        """
        self.__baseURL = baseURL.strip("/") and f"/{baseURL.strip('/')}"
        self.__handlersLocation = handlersLocation
        self.__routes = []
        self.__handlers = []
        # The following regular expression describes two capturing groups from the incoming request path:
        # 1) the DIRAC setup
        setup = r"[\w-]*"
        # 2) the DIRAC user group
        group = r"[\w.-]*"
        # E.g.: /<base URL>/s:<setup name>/g:<group name>/<your reuqest>..
        self.__setupGroupRE = f"/?(?:s:({setup})/?)?(?:g:({group})/?)?"
        # Describes the same, only as non-capturing groups, see `(?:...)` matcher.
        self.__shySetupGroupRE = f"/?(?:s:(?:{setup}/?)g:(?:{group})/?)?"
        self.log = gLogger.getSubLogger("Routing")

    def getPaths(self, dirName):
        """Get lists of paths for all installed and enabled extensions

        :param str dirName: path to handlers

        :return: list
        """
        pathList = []
        for extName in extensionsByPriority():
            metadata = getExtensionMetadata(extName)
            pathList.extend(map(str, metadata.get("web_resources", {}).get(dirName, [])))
        return pathList

    def __calculateRoutes(self):
        """Load all handlers and generate the routes

        :return: S_OK()/S_ERROR()
        """
        ol = ObjectLoader()
        handlerList = []
        self.log.debug("Add handles from: %s", self.__handlersLocation)
        for parentClass in [WebHandler, _WebHandler]:
            result = ol.getObjects(self.__handlersLocation, parentClass=parentClass, recurse=True, continueOnError=True)
            if not result["OK"]:
                return result
            handlerList += list(result["Value"].items())
        self.__handlers = collections.OrderedDict(handlerList)

        # ['/opt/dirac/pro/WebAppExt/WebApp/static', ...]
        staticPaths = self.getPaths("static")
        self.log.verbose("Static paths found:\n - %s" % "\n - ".join(staticPaths))
        self.__routes = []

        # Add some standard paths for static files
        statDirectories = Conf.getStaticDirs()
        self.log.info("The following static directories are used: %s" % str(statDirectories))
        for staticDirectory in statDirectories:
            # Get real static files directory
            realStaticDirectory = f"{rootPath}/webRoot/www/{staticDirectory}"
            # Make sure the directory exists
            if not os.path.exists(realStaticDirectory):
                os.makedirs(realStaticDirectory)
            # Let's add routes
            pattern = f"/{staticDirectory}/(.*)"
            self.__routes.append((pattern, StaticHandler, dict(pathList=[realStaticDirectory])))
            self.log.debug(" - Static route:", pattern)

        # Define the base URL of the request for static files
        static_base_url = self.__baseURL + self.__shySetupGroupRE

        for pattern in (r"/static/(.*)", r"/(favicon\.ico)", r"/(robots\.txt)"):
            self.__routes.append((static_base_url + pattern, StaticHandler, dict(pathList=staticPaths)))
            self.log.debug(" - Static route:", static_base_url + pattern)

        # Define the base URL of the request and add the setup group regex pattern
        handler_base_url = self.__baseURL + self.__setupGroupRE

        for hn in self.__handlers:
            self.log.info("Found handler %s" % hn)
            handler = self.__handlers[hn]

            # Set the base URL of the request
            handler.BASE_URL = handler_base_url

            # CHeck it has AUTH_PROPS or DEFAULT_AUTHORIZATION
            if isinstance(handler.AUTH_PROPS, type(None)) and isinstance(handler.DEFAULT_AUTHORIZATION, type(None)):
                return S_ERROR(f"Handler {hn} does not have AUTH_PROPS defined. Fix it!")

            # Generate tornado URLs
            self.__routes += handler._BaseRequestHandler__pre_initialize()

        return S_OK()

    def getHandlers(self):
        """Get handlers

        :return: S_OK()/S_ERROR()
        """
        if not self.__handlers:
            result = self.__calculateRoutes()
            if not result["OK"]:
                return result
        return S_OK(self.__handlers)

    def getRoutes(self):
        """Get routes

        :return: S_OK()/S_ERROR()
        """
        if not self.__routes:
            result = self.__calculateRoutes()
            if not result["OK"]:
                return result
        return S_OK(self.__routes)



import os
from tornado.web import StaticFileHandler, HTTPError

from DIRAC import rootPath


class StaticHandler(StaticFileHandler):
    def initialize(self, pathList, default_filename=None):
        # pathList:  ['/opt/dirac/pro/WebAppExt/WebApp/static', ...]
        self.pathList = [os.path.abspath(path) + os.path.sep for path in pathList]
        self.default_filename = default_filename
        self.root = rootPath

    def parse_url_path(self, url_path):
        if os.path.sep != "/":
            url_path = url_path.replace("/", os.path.sep)
        for possiblePath in self.pathList:
            possiblePath = os.path.join(possiblePath, url_path)
            if self.default_filename and os.path.isdir(possiblePath):
                possiblePath = os.path.join(possiblePath, self.default_filename)
            if os.path.isfile(possiblePath):
                return possiblePath
        raise HTTPError(404)



import collections
import json

from DIRAC import gLogger
from DIRAC.Core.Utilities import Time
from DIRAC.ResourceStatusSystem.Client.PublisherClient import PublisherClient
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import RSSMachine

from WebAppDIRAC.Lib.WebHandler import WebHandler, asyncGen


class SummaryHandlerMix(WebHandler):
    AUTH_PROPS = "all"
    ELEMENT_TYPE = None

    def _getSelectionData(self):
        """It returns the possible selection data"""
        callback = {"name": set(), "elementType": set(), "status": set(), "statusType": set(), "tokenOwner": set()}

        pub = PublisherClient()

        gLogger.info("Arguments to web_getSelectionData", repr(self.request.arguments))
        elementStatuses = pub.getElementStatuses(self.ELEMENT_TYPE, None, None, None, None, None)

        if elementStatuses["OK"]:

            for elementStatus in elementStatuses["Value"]:
                callback["status"].add(elementStatus[2])
                callback["name"].add(elementStatus[0])
                callback["elementType"].add(elementStatus[6])
                callback["statusType"].add(elementStatus[1])
                callback["tokenOwner"].add(elementStatus[8])

        for key, value in callback.items():

            callback[key] = sorted([[item] for item in list(value)])
            callback[key] = [["All"]] + callback[key]

        return callback

    def combine(self, elementValues):

        statuses = [element["Status"] for element in elementValues]

        statusSet = set(statuses)

        if len(statusSet) == 1:
            status = statusSet.pop()
            reason = "All %s" % status

        else:

            if set(["Active", "Degraded"]) & set(statusSet):
                status = "Degraded"
                reason = "Not completely active"

            else:
                status = "Banned"
                reason = "Not usable"

        #      if set( [ 'Unknown','Active', 'Degraded' ] ) & set( statusSet ):
        #        for upStatus in [ 'Active', 'Degraded' ]:
        #          if upStatus in statusSet:
        #            status = upStatus
        #            reason = '%d %s' % ( statuses.count( upStatus ), upStatus )
        #            break
        #      else:
        #        for downStatus in [ 'Unknown','Probing','Banned','Error' ]:
        #          if downStatus in statusSet:
        #            status = downStatus
        #            reason = '%d %s' % ( statuses.count( downStatus ), downStatus )
        #            break

        # Make a copy
        combined = {}
        combined.update(elementValues[0])
        combined["StatusType"] = "%d elements" % len(statuses)
        combined["Status"] = status
        combined["Reason"] = reason
        combined["DateEffective"] = ""
        combined["LastCheckTime"] = ""
        combined["TokenOwner"] = ""
        combined["TokenExpiration"] = ""

        return combined

    def _expand(self):
        """
        This method handles the POST requests
        """

        requestParams = self._requestParams()
        gLogger.info(requestParams)

        pub = PublisherClient()

        elements = pub.getElementStatuses(self.ELEMENT_TYPE, requestParams["name"], None, None, None, None)
        if not elements["OK"]:
            return {"success": "false", "error": elements["Message"]}

        elementList = [dict(zip(elements["Columns"], element)) for element in elements["Value"]]
        for element in elementList:
            element["DateEffective"] = str(element["DateEffective"])
            element["LastCheckTime"] = str(element["LastCheckTime"])
            element["TokenExpiration"] = str(element["TokenExpiration"])

        return {"success": "true", "result": elementList, "total": len(elementList)}

    def _action(self):

        requestParams = self._requestParams()
        if "action" in requestParams and requestParams["action"]:

            # pylint does not understand the action entry is not None any more
            actionName = requestParams["action"][0]  # pylint: disable=unsubscriptable-object

            methodName = actionName
            if not actionName.startswith("set"):
                methodName = "_get%s" % actionName

            try:
                return getattr(self, methodName)(requestParams)
            except AttributeError:
                return {"success": "false", "error": "bad action %s" % actionName}
        else:
            return {"success": "false", "error": "Missing action"}

    def setToken(self, requestParams):
        username = self.getUserName()

        if username == "anonymous":
            self.finish({"success": "false", "error": "Cannot perform this operation as anonymous"})
            return
        if "SiteManager" not in self.getProperties():
            self.finish({"success": "false", "error": "Not authorized"})
            return

        pub = PublisherClient()
        res = yield self.threadTask(
            pub.setToken,
            self.ELEMENT_TYPE,
            str(requestParams["name"][0]),
            str(requestParams["statusType"][0]),
            str(requestParams["status"][0]),
            str(requestParams["elementType"][0]),
            username,
            str(requestParams["lastCheckTime"][0]),
        )

        if not res["OK"]:
            self.finish({"success": "false", "error": res["Message"]})
        else:
            self.finish({"success": "true", "result": res["Value"]})

    def setStatus(self, requestParams):
        username = self.getUserName()

        if username == "anonymous":
            self.finish({"success": "false", "error": "Cannot perform this operation as anonymous"})
            return
        if "SiteManager" not in self.getProperties():
            self.finish({"success": "false", "error": "Not authorized"})
            return

        pub = PublisherClient()

        res = yield self.threadTask(
            pub.setStatus,
            self.ELEMENT_TYPE,
            str(requestParams["name"][0]),
            str(requestParams["statusType"][0]),
            str(requestParams["status"][0]),
            str(requestParams["elementType"][0]),
            username,
            str(requestParams["lastCheckTime"][0]),
        )

        if not res["OK"]:
            self.finish({"success": "false", "error": res["Message"]})
        else:
            self.finish({"success": "true", "result": res["Value"]})

    def _getHistory(self, requestParams):

        # Sanitize
        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "elementType" not in requestParams or not requestParams["elementType"]:
            self.finish({"success": "false", "error": "Missing elementType"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})

        pub = PublisherClient()
        res = yield self.threadTask(
            pub.getElementHistory,
            self.ELEMENT_TYPE,
            requestParams["name"],
            requestParams["elementType"],
            requestParams["statusType"],
        )

        if not res["OK"]:
            gLogger.error(res["Message"])
            self.finish({"success": "false", "error": "error getting history"})

        history = [[r[0], str(r[1]), r[2]] for r in res["Value"]]

        self.finish({"success": "true", "result": history, "total": len(history)})

    def _getPolicies(self, requestParams):

        # Sanitize
        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})

        pub = PublisherClient()
        res = yield self.threadTask(
            pub.getElementPolicies, self.ELEMENT_TYPE, requestParams["name"], requestParams["statusType"]
        )

        if not res["OK"]:
            gLogger.error(res["Message"])
            self.finish({"success": "false", "error": "error getting policies"})

        policies = [[r[0], r[1], str(r[2]), str(r[3]), r[4]] for r in res["Value"]]

        self.finish({"success": "true", "result": policies, "total": len(policies)})

    def _getDowntime(self, requestParams):

        # Sanitize
        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "elementType" not in requestParams or not requestParams["elementType"]:
            self.finish({"success": "false", "error": "Missing elementType"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})
        if "element" not in requestParams or not requestParams["element"]:
            self.finish({"success": "false", "error": "Missing element"})

        pub = PublisherClient()

        res = yield self.threadTask(
            pub.getDowntimes,
            str(requestParams["element"][-1]),
            str(requestParams["elementType"][-1]),
            str(requestParams["name"][-1]),
        )
        if not res["OK"]:
            gLogger.error(res["Message"])
            self.finish({"success": "false", "error": "error getting downtimes"})

        downtimes = [[str(dt[0]), str(dt[1]), dt[2], dt[3], dt[4]] for dt in res["Value"]]

        self.finish({"success": "true", "result": downtimes, "total": len(downtimes)})

    def _getTimeline(self, requestParams):

        # Sanitize
        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "elementType" not in requestParams or not requestParams["elementType"]:
            self.finish({"success": "false", "error": "Missing elementType"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})

        pub = PublisherClient()

        res = yield self.threadTask(
            pub.getElementHistory,
            self.ELEMENT_TYPE,
            str(requestParams["name"][-1]),
            str(requestParams["elementType"][-1]),
            str(requestParams["statusType"][-1]),
        )

        if not res["OK"]:
            gLogger.error(res["Message"])
            self.finish({"success": "false", "error": "error getting history"})

        history = []

        for status, dateEffective, reason in res["Value"]:

            # history.append( [ history[ -1 ][ 0 ], str( dateEffective - timedelta( seconds = 1 ) ), '' ] )

            history.append([status, str(dateEffective), reason])

        self.finish({"success": "true", "result": history, "total": len(history)})

    def _getTree(self, requestParams):

        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "elementType" not in requestParams or not requestParams["elementType"]:
            self.finish({"success": "false", "error": "Missing elementType"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})

        pub = PublisherClient()

        res = yield self.threadTask(pub.getTree, str(requestParams["elementType"][-1]), str(requestParams["name"][-1]))
        if not res["OK"]:
            gLogger.error(res["Message"])
            self.finish({"success": "false", "error": "error getting tree"})
        res = res["Value"]

        siteName = list(res)[0]

        tree = [[siteName, None, None, None]]
        for k, v in res[siteName]["statusTypes"].items():
            tree.append([None, k, v, siteName])

        tree.append(["ces", None, None, siteName])
        for ce, ceDict in res[siteName]["ces"].items():
            tree.append([ce, None, None, "ces"])
            for k, v in ceDict.items():
                tree.append([None, k, v, ce])

        tree.append(["ses", None, None, siteName])
        for se, seDict in res[siteName]["ses"].items():
            tree.append([se, None, None, "ses"])
            for k, v in seDict.items():
                tree.append([None, k, v, se])

        self.finish({"success": "true", "result": tree, "total": len(tree)})

    def _getInfo(self, requestParams):
        if "name" not in requestParams or not requestParams["name"]:
            self.finish({"success": "false", "error": "Missing name"})
        if "elementType" not in requestParams or not requestParams["elementType"]:
            self.finish({"success": "false", "error": "Missing elementType"})
        if "statusType" not in requestParams or not requestParams["statusType"]:
            self.finish({"success": "false", "error": "Missing statusType"})
        if "element" not in requestParams or not requestParams["element"]:
            self.finish({"success": "false", "error": "Missing element"})

        pub = PublisherClient()

        res = yield self.threadTask(
            pub.getElementStatuses,
            str(requestParams["element"][-1]),
            str(requestParams["name"][-1]),
            str(requestParams["elementType"][-1]),
            str(requestParams["statusType"][-1]),
            None,
            None,
        )

        if not res["OK"]:
            self.finish({"success": "false", "error": res["Message"]})
        elif not res["Value"]:
            self.finish({"success": "false", "error": "Nothing found."})

        else:

            columns = res["Columns"]

            res = dict(zip(columns, res["Value"][0]))
            res["DateEffective"] = str(res["DateEffective"])
            res["LastCheckTime"] = str(res["LastCheckTime"])
            res["TokenExpiration"] = str(res["TokenExpiration"])

            self.finish({"success": "true", "result": res, "total": len(res)})

    def _requestParams(self):
        """
        We receive the request and we parse it, in this case, we are doing nothing,
        but it can be certainly more complex.
        """
        gLogger.always("!!!  PARAMS: ", repr(self.request.arguments))

        responseParams = {
            "element": None,
            "name": None,
            "elementType": None,
            "statusType": None,
            "status": None,
            "tokenOwner": None,
            "lastCheckTime": None,
            "action": None,
        }

        for key in responseParams:
            value = self.get_argument(key, "")
            if value:
                responseParams[key] = list(json.loads(value))

        return responseParams


class ResourceSummaryHandler(SummaryHandlerMix):

    ELEMENT_TYPE = "Resource"

    @asyncGen
    def web_getSelectionData(self):
        callback = yield self.threadTask(self._getSelectionData)
        self.finish(callback)

    @asyncGen
    def web_expand(self):
        callback = yield self.threadTask(self._expand)
        self.finish(callback)

    @asyncGen
    def web_action(self):
        callback = yield self.threadTask(self._action)
        self.finish(callback)

    @asyncGen
    def web_getResourceSummaryData(self):
        """This method returns the data required to fill the grid."""

        requestParams = self._requestParams()
        gLogger.info(requestParams)

        pub = PublisherClient()

        elementStatuses = yield self.threadTask(
            pub.getElementStatuses,
            self.ELEMENT_TYPE,
            requestParams["name"],
            requestParams["elementType"],
            requestParams["statusType"],
            requestParams["status"],
            requestParams["tokenOwner"],
        )
        if not elementStatuses["OK"]:
            self.finish({"success": "false", "error": elementStatuses["Message"]})
            return

        elementTree = collections.defaultdict(list)

        for element in elementStatuses["Value"]:

            elementDict = dict(zip(elementStatuses["Columns"], element))

            elementDict["DateEffective"] = str(elementDict["DateEffective"])
            elementDict["LastCheckTime"] = str(elementDict["LastCheckTime"])
            elementDict["TokenExpiration"] = str(elementDict["TokenExpiration"])

            elementTree[elementDict["Name"]].append(elementDict)

        elementList = []

        for elementValues in elementTree.values():

            if len(elementValues) == 1:
                elementList.append(elementValues[0])
            else:

                elementList.append(self.combine(elementValues))

        rssMachine = RSSMachine(None)

        yield self.threadTask(rssMachine.orderPolicyResults, elementList)

        timestamp = Time.dateTime().strftime("%Y-%m-%d %H:%M [UTC]")

        self.finish({"success": "true", "result": elementList, "total": len(elementList), "date": timestamp})
